// iteratees in C proof-of-concept
// pesco 2013, 2016

// cf. http://okmij.org/ftp/Streams.html#iteratee

// NOTES:
//
// allocates A LOT of stuff mainly because iteratees, or rather their
// associated environment structs, are treated as immutable values. rationale
// was that it makes things look more like they are in Haskell for now.
// building a proof-of-concept around mutable iteratees is TODO.
//
// the other way out is to construct an efficient bespoke garbage collector for
// the iteratee env's (high-churn, small objects).
//
// benchmarking the test4 (word count) example, the proof-of-concept garbage
// collector as implemented has negligable impact on performance. with arena
// size set to 1MB, collecting on every grow(), the test case, a file of ~14M
// words, peaks at 4 concurrent arenas with about 3MB in use. overall runtime
// was on par with, even slightly below, that of 'wc -w' under LANG=C (i.e. no
// multibyte decoding).

// TODO / further directions:
//
// * adapt allocator for memory-constrained applications
// * enumerator/enumeratee composition
// * parallel composition of iteratees (feed same stream into multiple); could
//   implement 'wc -c -w -l' then


#include <stdio.h>
#include <stdlib.h>     // exit()
#include <string.h>     // memcpy()
#include <stdarg.h>     // va_*()
#include <stdint.h>     // uintptr_t
#include <assert.h>
#include <errno.h>
#include "iter.h"
#include "gc.h"

#define BUFLEN 4096             // size of input buffer


#define not_reached() do { \
    fprintf(stderr, "%s:%d: assertion failed: code reached\n", \
                    __FILE__, __LINE__); \
    exit(1); \
} while(0)


// running iteratees

void *finish(Iteratee iter)
{
    Stream end = {.state = END, .error = NULL};

    switch(iter.state) {
    case DONE:
        return iter.result;
    case CONT:
        iter = iter.cont(iter.env, &end);
        assert(iter.state != CONT);
        return finish(iter);
    case STOP:
        // XXX error reporting doesn't belong here
        fprintf(stderr, "iteratee error: %s\n", iter.error);
        return NULL;
    default:
        not_reached();
    }
}

// we can wrap arbitrary code in an iteratee and use 'finish' to run it.
// this is in fact Oleg's name for the function.
void *run(Iteratee iter)
{
    return finish(iter);
}


// building iteratees

Iteratee done(void *result)
{
    Iteratee iter = {.state = DONE, .result = result};
    return iter;
}

Iteratee cont(Iteratee (*f)(void *, Stream *), void *env)
{
    Iteratee iter = {.state = CONT, .cont = f, .env = env, .error = NULL};
    return iter;
}

Iteratee stop(const char *error, Iteratee (*resume)(void *, Stream *), void *env)
{
    Iteratee iter = {.state = STOP, .error = error, .cont = resume, .env = env};
    return iter;
}

Iteratee fail(const char *error);
Iteratee f_fail(void *env, Stream *s)
{
    return fail((const char *)env);
}
Iteratee fail(const char *error)
{
    return stop(error, f_fail, (void *)error);
}

struct Env_bind {
    Iteratee iter;
    Iteratee (*k)(void *env, void *result);
    void *k_env;
};
static Iteratee f_bind(void *env_, Stream *s)
{
    struct Env_bind *env = env_;
    const char *error;
    Iteratee it;

    switch(env->iter.state) {
    case DONE:
        it = env->k(env->k_env, env->iter.result);
        if(it.state == CONT)
            it = it.cont(it.env, s);
        return it;
    case CONT:
        // duplicate the environment
        env = DUP(env, struct Env_bind);
        // run left side as far as it will, updating env
        env->iter = env->iter.cont(env->iter.env, s);
        if(env->iter.state == CONT)
            return cont(f_bind, env);   // wait for more input
        else
            return f_bind(env, s);      // state change; reenter with new env
    case STOP:
        error = env->iter.error;
        // duplicate the environment
        env = NEW(struct Env_bind);
        *env = *(struct Env_bind *)env_;
        // build resume environment
        env->iter = cont(env->iter.cont, env->iter.env);
        return stop(error, f_bind, env);
    default:
        not_reached();
    }
}
Iteratee bind(Iteratee iter, Iteratee (*k)(void *env, void *result), void *env)
{
    struct Env_bind *env_bind = NEW(struct Env_bind);
    env_bind->iter = iter;
    env_bind->k = k;
    env_bind->k_env = env;
    return cont(f_bind, env_bind);
}

struct Env_bind_ {
    Iteratee i;
    Iteratee j;
};
static Iteratee f_bind_(void *env_, Stream *s)
{
    struct Env_bind_ *env = env_;
    Iteratee it = env->i;

    for(;;) {
        // run left side as far as it will
        if(it.state == CONT) 
            it = it.cont(it.env, s);

        switch(it.state) {
        case DONE:
            // left side done; switch over
            if(env->j.state == CONT)
                return env->j.cont(env->j.env, s);
            else
                return env->j;
        case CONT:
            // duplicate and update the environment
            env = NEW(struct Env_bind_);
            env->i = it;
            env->j = ((struct Env_bind_ *)env_)->j;

            return cont(f_bind_, env);  // wait for more input
        case STOP:
            // build resume environment
            env = NEW(struct Env_bind_);
            env->i = cont(it.cont, it.env);
            env->j = ((struct Env_bind_ *)env_)->j;

            return stop(it.error, f_bind_, env);
        }
    }
    not_reached();
}
Iteratee bind_(Iteratee i, Iteratee j)
{
    struct Env_bind_ *env = NEW(struct Env_bind_);
    env->i = i;
    env->j = j;
    return cont(f_bind_, env);
}

struct Env_seq {
    Iteratee *iters;
    size_t n;

    size_t i;           // current position, for stopped sequences
    void **results;     // current results
};
static Iteratee f_seq(void *env_, Stream *s)
{
    struct Env_seq *env = env_, *env_orig = env_;

    Iteratee iter = env->iters[env->i];
    while(env->i < env->n) {
        switch(iter.state) {
        case DONE:
            // save result; continue with next iteratee
            if(env == env_orig)
                env = DUP(env, struct Env_seq);
            if(env->results) {
                if(env->results == env_orig->results)
                    env->results = DUPA(env->results, void *, env->n);
                env->results[env->i] = iter.result;
            } else if(env->i+1 == env->n) {
                // final result
                // yes, passing it through 'results' (plural) is a bit hacky
                env->results = iter.result;
            }
            env->i++;
            if(env->i < env->n)
                iter = env->iters[env->i];
            break;
        case CONT:
            // feed stream; repeat
            iter = iter.cont(iter.env, s);
            if(iter.state == CONT) {
                if(env == env_orig)
                    env = DUP(env, struct Env_seq);
                if(env->iters == env_orig->iters)
                    env->iters = DUPA(env->iters, Iteratee, env->n);
                env->iters[env->i] = iter;
                return cont(f_seq, env);
            }
            break;
        case STOP:
            // one iter stops -> sequence stops
            return stop(iter.error, f_seq, env);
        }
    }

    return done(env->results);
}
Iteratee seq(size_t n, ...)
{
    va_list ap;

    // allocate environment
    struct Env_seq *env = NEW(struct Env_seq);
    env->iters = NEWA(Iteratee, n);
    env->i = 0;
    env->n = n;
    env->results = NEWA(void *, n);

    // save argument list in environment
    va_start(ap, n);
    for(size_t i=0; i<n; i++)
        env->iters[i] = va_arg(ap, Iteratee);
    va_end(ap);

    return cont(f_seq, env);
}
Iteratee seq_(size_t n, ...)
{
    va_list ap;

    // allocate environment
    struct Env_seq *env = NEW(struct Env_seq);
    env->iters = NEWA(Iteratee, n);
    env->i = 0;
    env->n = n;
    env->results = NULL;    // only return the last result

    // save argument list in environment
    va_start(ap, n);
    for(size_t i=0; i<n; i++)
        env->iters[i] = va_arg(ap, Iteratee);
    va_end(ap);

    return cont(f_seq, env);
}


// primitive iteratees

static Iteratee f_head(void *env, Stream *s)
{
    switch(s->state) {
    case END:
        return stop(s->error? s->error : "unexpected EOF", f_head, env);
    case CHUNK:
        if(s->n > 0) {
            // consume stream element
            void *result = new(s->elemsize);
            memcpy(result, s->elems, s->elemsize);
            s->n--;
            s->elems += s->elemsize;
            // return result
            return done(result);
        } else {
            return cont(f_head, env);   // no input available, keep waiting
        }
    default:
        not_reached();
    }
}
const Iteratee head = {.state = CONT, .cont = f_head, .env = NULL};

static Iteratee f_chunk(void *env, Stream *s)
{
    switch(s->state) {
    case END:
        return stop(s->error? s->error : "unexpected EOF", f_chunk, env);
    case CHUNK:
        s->n = 0;
        return done(s);
    default:
        not_reached();
    }
}
const Iteratee chunk = {.state = CONT, .cont = f_chunk, .env = NULL};

static Iteratee f_count(void *env, Stream *s)
{
    uintptr_t n = (uintptr_t)env;
    switch(s->state) {
    case END:
        return done((void *)n);
    case CHUNK:
        n += s->n;
        s->n = 0;
        return cont(f_count, (void *)n);
    default:
        not_reached();
    }
}
const Iteratee count = {.state = CONT, .cont = f_count, .env = 0};

// this one is like count, except it doesn't even count
static Iteratee f_drop(void *env, Stream *s)
{
    switch(s->state) {
    case END:
        return done(NULL);
    case CHUNK:
        s->n = 0;
        return cont(f_drop, NULL);
    default:
        not_reached();
    }
}
const Iteratee drop = {.state = CONT, .cont = f_drop, .env = NULL};

// concatenate input into one buffer (returned as a Stream structure)
static Iteratee f_take(void *env, Stream *s)
{
    Stream *r, *t = env;
    size_t n = t? t->n : 0;             // number of elements in t

    switch(s->state) {
    case END:
        return done(t);
    case CHUNK:
        // make new result
        r = NEW(Stream);
        r->state = CHUNK;               // always CHUNK
        r->elemsize = s->elemsize;
        r->n = n + s->n;
        r->elems = r->orig = new(r->elemsize * r->n);

        // copy old partial result over
        if(t) {
            assert(s->elemsize == t->elemsize);
            memcpy(r->elems, t->elems, n);
        }

        // append new input
        memcpy(r->elems + n, s->elems, s->n);
        s->n = 0;   // consume

        return cont(f_take, r);
    default:
        not_reached();
    }
}
const Iteratee take = {.state = CONT, .cont = f_take, .env = NULL};

// a variation of take that requires at least one element of input
static Iteratee f_take1(void *env, Stream *s)
{
    switch(s->state) {
    case END:
        return stop("EOF", f_take1, env);
    case CHUNK:
        if(s->n < 1)
            return cont(f_take1, env);
        else
            return f_take(env, s);
    default:
        not_reached();
    }
}
const Iteratee take1 = {.state = CONT, .cont = f_take1, .env = NULL};

// a variation of drop that requires at least one element of input
static Iteratee f_drop1(void *env, Stream *s)
{
    switch(s->state) {
    case END:
        return stop("EOF", f_drop1, env);
    case CHUNK:
        if(s->n < 1)
            return cont(f_drop1, env);
        else
            return f_drop(env, s);
    default:
        not_reached();
    }
}
const Iteratee drop1 = {.state = CONT, .cont = f_drop1, .env = NULL};


// enumerators

Enumerator enumerator(Iteratee (*run)(void *, Iteratee), void *env)
{
    Enumerator ret = {.run = run, .env = env};
    return ret;
}

Iteratee apply(Enumerator en, Iteratee iter)
{
    return en.run(en.env, iter);
}

struct Env_append {
    Enumerator a;
    Enumerator b;
};
Iteratee f_append(void *env_, Iteratee iter) {
    struct Env_append *env = env_;
    iter = apply(env->a, iter);
    return apply(env->b, iter);
}
Enumerator append(Enumerator a, Enumerator b)
{
    struct Env_append *env = NEW(struct Env_append);
    env->a = a;
    env->b = b;
    return enumerator(f_append, env);
}

Iteratee f_enumf(void *env, Iteratee iter)
{
    FILE *f = env;
    uint8_t buf[BUFLEN];
    size_t n;

    Stream s;
    s.elemsize = 1;
    while(iter.state == CONT) {
        n = fread(buf, 1, BUFLEN, f);
        if(n == 0) {
            break;
        } else {
            s.state = CHUNK;
            s.elems = buf;
            s.orig = buf;
            s.n = n;
        }
        iter = iter.cont(iter.env, &s);
        if(n < BUFLEN)  // short read
            break;
    }
    if(iter.state == CONT && ferror(f)) {
        s.state = END;
        s.error = "read failed";
        iter = iter.cont(iter.env, &s);
    }

    return iter;
}
Enumerator enumf(FILE *f)
{
    return enumerator(f_enumf, f);
}

static Iteratee f_enumfile(void *env, Iteratee iter)
{
    const char *path = env;

    FILE *f = fopen(path, "r");
    if(!f) {
        const char *e_ = strerror(errno);
        size_t n = strlen(e_);
        char *e = NEWA(char, n+1);
        memcpy(e, e_, n);
        e[n] = '\0';
        return fail(e);
    }

    iter = f_enumf(f, iter);

    fclose(f);
    return iter;
}
Enumerator enumfile(const char *path)
{
    return enumerator(f_enumfile, (void *)path);
}


// enumeratees

struct Env_decode {
    Iteratee elem;          // could be an intermediate state
    Iteratee inner;
    Iteratee elem_orig;
};
Iteratee f_decode_outer(void *env_, Stream *s)
{
    struct Env_decode *env = env_;

    Iteratee elem = env->elem;
    Iteratee inner = env->inner;

    // inner input stream
    Stream is = {.state = CHUNK, .elems = NULL, .n = 0,
                 .elemsize = sizeof(void *), .orig = NULL};

    // handle outer EOF
    if(s->state == END) {
        // if elem is still waiting for input, feed EOF to it
        if(elem.state == CONT) {
            elem = elem.cont(elem.env, s);

            // if inner is waiting for input, feed result to it
            if(inner.state == CONT && elem.state == DONE) {
                is.elems = is.orig = &elem.result;
                is.n = 1;
                inner = inner.cont(inner.env, &is);
            }
        }

        // inner iteratee ~> outer result
        Iteratee *result = DUP(&inner, Iteratee);
        return done(result);
    }

    void *buf[BUFLEN];
    is.elems = is.orig = buf;

    while(inner.state == CONT) {
        // generate inner input elements
        is.n = 0;
        while(elem.state == CONT && s->n > 0 && is.n < BUFLEN) {
            // feed outer input to elem iter
            elem = elem.cont(elem.env, s);

            if(elem.state != DONE)
                break;

            // add result to inner input and consume the next element
            buf[is.n++] = elem.result;
            elem = env->elem_orig;
        }

        if(is.n == 0) {
            // save current inner state and (partial) elem parse
            env = DUP(env, struct Env_decode);
            env->inner = inner;
            env->elem = elem;
        
            // ask for more outer input
            return cont(f_decode_outer, env);
        }

        // feed input to inner iteratee
        inner = inner.cont(inner.env, &is);
    }

    // XXX should a STOP on inner become a STOP on outer?

    Iteratee *result = DUP(&inner, Iteratee);
    return done(result);
}
Iteratee f_decode(void *env, Iteratee iter)
{
    Iteratee *elemiter = env;

    // return an iteratee that consumes elements via elemiter, feeds them
    // to the (inner) iter, and returns the "remaining" inner iter.

    struct Env_decode *envi = NEW(struct Env_decode);
    envi->elem = envi->elem_orig = *elemiter;
    envi->inner = iter;
    return cont(f_decode_outer, envi);
}
Enumeratee decode(Iteratee elemiter)
{
    Iteratee *env = NEW(Iteratee);
    *env = elemiter;
    return enumerator(f_decode, env);
}

// pass only elements that satisfy a predicate
struct Env_filter {
    int (*pred)(void *elem);
    Iteratee inner;
};
static Iteratee f_filter_(void *env_, Stream *s)
{
    struct Env_filter *env = env_;

    Iteratee inner = env->inner;
    size_t size = s->elemsize;

    // inner input stream
    Stream is = {.state = CHUNK, .elems = NULL, .n = 0,
                 .elemsize = size, .orig = NULL};

    // handle outer EOF
    if(s->state == END) {
        // inner iteratee ~> outer result
        Iteratee *result = DUP(&inner, Iteratee);
        return done(result);
    }

    void *buf[BUFLEN];
    is.elems = is.orig = buf;

    while(inner.state == CONT) {
        // generate inner input elements
        is.n = 0;
        while(s->n > 0 && is.n < BUFLEN/size) {
            if(env->pred(s->elems)) {
                // add element to inner input
                memcpy(buf + (is.n * size), s->elems, size);
                is.n++;
                s->n--; s->elems += size;
            }
        }

        if(is.n == 0) {
            // save current inner state
            env = DUP(env, struct Env_filter);
            env->inner = inner;
        
            // ask for more outer input
            return cont(f_filter_, env);
        }

        // feed input to inner iteratee
        inner = inner.cont(inner.env, &is);
    }

    Iteratee *result = DUP(&inner, Iteratee);
    return done(result);
    
}
static Iteratee f_filter(void *pred_, Iteratee iter)
{
    int (*pred)(void *elem) = pred_;

    // return an iteratee that consumes elements, tests them with pred,
    // feeds the matching ones to iter and yields the remaining iter.

    struct Env_filter *env = NEW(struct Env_filter);
    env->pred = pred;
    env->inner = iter;
    return cont(f_filter_, env);
}
Enumeratee filter(int (*pred)(void *elem))
{
    return enumerator(f_filter, pred);
}

// pass elements while or until they satisfy a predicate
struct Env_while {
    int (*pred)(void *elem);
    int invert;                 // 0=while 1=until
    Iteratee inner;
};
static Iteratee f_while_(void *env_, Stream *s)
{
    struct Env_while *env = env_;

    Iteratee inner = env->inner;
    size_t size = s->elemsize;

    // inner input stream
    Stream is = {.state = CHUNK, .elems = NULL, .n = 0,
                 .elemsize = size, .orig = NULL};

    // handle outer EOF
    if(s->state == END) {
        // inner iteratee ~> outer result
        Iteratee *result = DUP(&inner, Iteratee);
        return done(result);
    }

    void *buf[BUFLEN];
    is.elems = is.orig = buf;

    int stop = 0;
    while(inner.state == CONT && !stop) {
        // generate inner input elements
        is.n = 0;
        while(s->n > 0 && is.n < BUFLEN/size) {
            stop = !env->pred(s->elems);
            if(env->invert)
                stop = !stop;
            if(stop)
                break;

            // add element to inner input
            memcpy(buf + (is.n * size), s->elems, size);
            is.n++;
            s->n--; s->elems += size;
        }

        if(is.n == 0) {
            if(!stop) {
                // save current inner state
                env = DUP(env, struct Env_while);
                env->inner = inner;
            
                // ask for more outer input
                return cont(f_while_, env);
            }
        } else {
            // feed input to inner iteratee
            inner = inner.cont(inner.env, &is);
        }
    }

    Iteratee *result = DUP(&inner, Iteratee);
    return done(result);
}
struct Env_ewhile {
    int (*pred)(void *elem);
    int invert;
};
static Iteratee f_while(void *env_, Iteratee iter)
{
    struct Env_ewhile *env = env_;

    struct Env_while *envi = NEW(struct Env_while);
    envi->pred = env->pred;
    envi->inner = iter;
    envi->invert = env->invert;
    return cont(f_while_, envi);
}
Enumeratee ewhile(int (*pred)(void *elem))
{
    struct Env_ewhile *env = NEW(struct Env_ewhile);
    env->pred = pred;
    env->invert = 0;
    return enumerator(f_while, env);
}
Enumeratee until(int (*pred)(void *elem))
{
    struct Env_ewhile *env = NEW(struct Env_ewhile);
    env->pred = pred;
    env->invert = 1;
    return enumerator(f_while, env);
}

static Iteratee f_fuse(void *env, void *v)
{
    Iteratee inner = *(Iteratee *)v;
    Stream end = {.state = END, .error = NULL};

    if(inner.state == CONT)
        inner = inner.cont(inner.env, &end);
    assert(inner.state != CONT);

    switch(inner.state) {
    case DONE:
        return done(inner.result);
    case STOP:
        return fail(inner.error);  // XXX resume?
    default:
        not_reached();
    }
}
Iteratee fuse(Iteratee ie)
{
    return bind(ie, f_fuse, NULL);
}

Iteratee wrap(Enumeratee ee, Iteratee ie)
{
    return fuse(apply(ee, ie));
}


